/*
** Simple example for a monitor with pthreads.
**
** Compile with:
**    g++ -Wall -o monitor -pthread monitor.c
*/

#include <unistd.h>
#include <iostream>
#include <thread>
#include <mutex>
#include <condition_variable>

/* Declare and initialize the mutex */
std::mutex mutex;

/* Declare and initialize the condition variable */
std::condition_variable cond;

/* Indicates the the result is valid */
volatile int ready = 0;

/* Buffer space for a result of type int */
volatile int result;

void StoreResult(int arg) 
{
	/* Lock the mutex by creating a new 'unique_lock' object.
	   The mutex is automatically unlocked when this object
	   is destroyed, i.e. when the method returns. */
	std::unique_lock<std::mutex> lock(mutex);
	result = arg;  /* store result */
	ready = 1;
	cond.notify_all();
}

int ReadResult()
{
	/* Lock the mutex by creating a new 'unique_lock' object.
	   The mutex is automatically unlocked when this object
	   is destroyed, i.e. when the method returns. */
   std::unique_lock<std::mutex> lock(mutex);
   while (ready != 1)
      cond.wait(lock);
   return result; /* read and return result */
}

/*
** Producer thread: sleeps 3 seconds and creates result.
*/
void Producer()
{
   std::cout << "Producer started\n";
   sleep(3); /* wait 3 seconds ... */
   std::cout << "Producer stores result\n";
   StoreResult(4711);
}

/*
** Consumer thread: when it tries to read the result, it will be
** blocked until the result is ready.
*/
void Consumer()
{
   std::cout << "Consumer started\n";
   std::cout << "Result: " << ReadResult() << "\n";
}

int main(int argc, char **argv)
{
  /* Create threads */
  std::thread t1(Consumer);
  std::thread t2(Producer);

  /* Wait for threads to finish */
  t1.join();
  t2.join();

  return 0;
}

