#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <sys/time.h>

using namespace std;

/*
** Quicksort: sort elements 'lo' ... 'hi' in array 'a'.
*/
void quicksort(int *a, int lo, int hi)
{
	int i, j;
	int pivot;

	if (lo >= hi) {
		/*
		** Subarray is empty or has just one element, i.e., it is already sorted
		*/
		return;
	}

	/* Determine pivot element */
	pivot = a[(lo+hi)/2];

	/* Split a into two parts */
	i = lo;
	j = hi;
	while (i <= j) {    
		while (a[i] < pivot)
			i++; 
		while (a[j] > pivot)
			j--;
		if (i <= j) {
			/*
			** swap a[i] and a[j]
			*/
			int tmp = a[i];
			a[i] = a[j];
			a[j] = tmp;

			i++;
			j--;
		}
	}
	
	/* Recursion */
    /* firstprivate(a,lo,j) can be omitted here, since in this case,
	   firstprivate is the default. See p. 79 of the OpenMP 3.0 standard. */
#pragma omp task firstprivate(a,lo,j) if (j-lo > 10000)
	quicksort(a, lo, j);
	quicksort(a, i, hi);
	/* If this directive is removed, quicksort() returns before the array
	   is fully sorted. This may lead to problems, when the parallel region
	   extends after the initial call of quicksort(), as it is the case below.
	*/
#pragma omp taskwait
}

/*
** Initialize an array 'a' of length 'n' with random numbers
*/
void initialize(int *a, int n)
{
	int i;
	
	for (i = 0; i < n; i++) {
		a[i] = rand();
	}
}

/*
** Check whether an array 'a' of length 'n' is sorted.
** If so, return 'true', else 'false'.
*/
int isSorted(int *a, int n)
{
	int i;
	
	for (i = 1; i < n; i++) {
		if (a[i] < a[i-1])
			return false;
	}
	return true;
}

/*
** Auxiliary function: returns the current time in seconds as a floating point number.
*/
double getTime()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_sec + tv.tv_usec * 0.000001;
}

int main(int argc, char **argv)
{
	int *array;
	int n;
	double timeStart, timeEnd;

	if (argc < 2) {
		cerr << "Usage: qsort <array-size>\n";
		return 1;
	}

	/*
	** 'n' is the command line argument, converted into an integer number
	*/
	n = atoi(argv[1]);
	if ((n <= 0) || (n > 50000000)) {
		cerr << "Illegal array size!\n";
		return 2;
	}

	/*
	** Allocate the array
	*/
	array = new int[n];

	/*
	** From here on, we have multiple threads
	*/
#pragma omp parallel
#pragma omp single nowait /* This directive ensures that only one of the threads
							 executes the block.. */
	{
		/*
		** Initialize the array
		*/
		initialize(array, n);

		/*
		** Sort and measure the time ...
		*/
		timeStart = getTime();
		quicksort(array, 0, n-1);
		timeEnd = getTime();

		/*
		** Check whether the result is sorted correctly.
		*/
		if (isSorted(array, n)) {
			cout << "OK, array is sorted!\n";
		}
		else {
			cout << "OOPS, array is NOT sorted!\n";
		}
	}

	cout << "Time for sorting: " << (timeEnd-timeStart) << " seconds\n";

	return 0;
}
